
   ####################################################################################
   ### 	R program for fitting Bayesian multivariate density deconvolution models  	### 
   ###		(last modified on Aug 03, 2016, Duke University, Durham, NC, USA)		###
   ####################################################################################   
   
   
   	#################
   	### Author(s) ###
   	#################
   
# Author of the codes: Abhra Sarkar (abhra.stat@gmail.com)
# Authors of the paper: Abhra Sarkar, Debdeep Pati, Antik Chakraborty, Bani K. Mallick, Raymond J. Carroll
# Author for correspondence: Raymond J. Carroll (carroll@stat.tamu.edu)

 
   	###################
   	### Description ###
   	###################
   
# The goal is to estimate the density f_{X} of the vector valued variable of interest \bX. 
# Precise observations on \bX are NOT available.
# Instead, replicated proxies contmainated by measurement errors \bU are generated by the model: 
# \bW_{ij} = \bX_{i} + \bU_{ij}, for i=1,2,...,n, and j=1,...,m_{i}.
#
# The density f_{X} is modeled using overfitted finite mxitures of multivariate normal kernels:
# 		f_{X} = \sum_{k=1}^{K} \pi_{k} \MVN_{d}(\bmu_{k},\bSigma_{k}).
#
# The measurement errors \bU are assumed to have the following structure: 
#		\bU = S(\bX)\bepsilon,
# where S(\bX) = \diag(s_{1}(X_{1}),...,s_{d}(X_{d})) models the conditional variability in the measurement errors.
# The scaled errors \bepsilon are assumed to be independent from \bX.
#  
# The density of the scaled errors is also modeled using finite mixtures of multivariate normal kernels: 
# 		f_{\epsilon} = \sum_{k=1}^{K} \pi_{k} \MVN_{d}(\bmu_{k},\bSigma_{k}),
# subject to the moment restriction:
#		\sum_{k=1}^{K} \pi_{k}\bmu_{k} = 0.
# 
# The variance functions s_{m}^{2}(X_{m}) are estimated by fitting reparametrized versions of the univariate submodels. 
# These estmates are then kept fixed while fitting the multivariate model.
#
# To induce sparsity, the models for both f_{X} and f_{\epsilon} use latent factor characterization of the covariance matrices: 
# 		\bSigma_{k} = \bLambda_{k}\bLambda_{k} + \bOmega,
# where \bOmega = \diag(\sigma_{1}^{2},...,\sigma_{d}^{2}) is a diagonal matrix common to all components.
# Sparsity inducing multivariate gamma process shrinkage priors are assigned to the elements of the factor loading matrix.






rm(list=ls())
seedno = 1
set.seed(seedno)

setwd("/Users/abhrasarkar/Dropbox/Projects with Dr. C/X ME Multivariate Density Deconvolution/Final Codes")


 	###########################################
 	### Source Files and Libraries Required ###
 	###########################################

source("BAYES_MVT_DECON_FUNCTS.R")
source("BAYES_UNIV_DECON_HTR.R")

library(mvtnorm)
library(mclust)
library(RColorBrewer)
library(MCMCpack)	# Dependencies : coda, lattice and MASS
library(corpcor)



#######################
### MCMC parameters ###
#######################

# iteration_no_to_change_on_the_fly must be set either at a value smaller than burnin, or at a value larger than simsize. 
# At iteration_no_to_change_on_the_fly, if set less than burnin, reduntant components are dumped.
simsize = 300
burnin = 100
iteration_no_to_change_on_the_fly = 80



	#################
	### Load Data ###
	#################


Data = load("Simulated_Data_Set.RData")

d = ncol(ws)

# For real data sets, it is recommended that the surrogates be scaled to be unitless and have common range. 
# This can be done by uncommenting lines 103-105 below.
# Non-linear transformations that do not preserve the additivity of the data generating model are not recommended.  
# The common range of values of X is specified by "range.xs". 
# Most real data sets will be highly right skewed. 
# So setting the range at 20 units may bring the effective range to much smaller values. 
# For the EATS data set analyzed in the main paper, for example, this brings the effective range to about 10 units. 

#range.xs = 20
#for(ii in 1:d)
#	ws[,ii] = range.xs*(ws[,ii]-min(ws[,ii]))/(max(ws[,ii])-min(ws[,ii])) 

mis = as.data.frame(table(inds))$Freq	# Number of replicates per individual
n.xs = n = length(unique(inds))			# Total no of individuals
inds = rep(1:n,times=mis)				# Individual IDs replicated to match the proxies
n.es = N = sum(mis)						# Total no of surrogates

inds1 = inds2 = numeric(n)
inds1[1] = 1
inds2[1] = inds1[1]+mis[1]-1
for(ii in 2:n)
	{
	inds1[ii] = inds1[ii-1]+mis[ii-1]
	inds2[ii] = inds1[ii] + mis[ii]-1
	}
	
xs = matrix(0,nrow=n,ncol=d)
for(ii in 1:n)
	xs[ii,] = colMeans(ws[inds1[ii]:inds2[ii],])
us = (ws - xs[inds,])



###########################################
### Prior Parameters and Initial Values ###
###########################################

inds1 = inds2 = numeric(n)
inds1[1] = 1
inds2[1] = inds1[1]+mis[1]-1
for(ii in 2:n)
	{
	inds1[ii] = inds1[ii-1]+mis[ii-1]
	inds2[ii] = inds1[ii] + mis[ii]-1
	}
	
xs = matrix(0,nrow=n,ncol=d)
K.t = 10
knots = matrix(0,nrow=K.t,ncol=d)
thetas = matrix(0,nrow=K.t+1,ncol=d)
es = matrix(0,nrow=sum(mis),ncol=d)


### Run Univariate Submodels ###
# This produces estimates of the variance functions that capture the conditional variability. 
# This also gives initial estimates of X, number of clusters etc. as important by-products.

simsize_univ = 3000
burnin_univ = 1000
for(jj in 1:d)
{
filename = paste("MVT_Decon_HTR_Initial_Values",jj,sep="_")
filename = paste(filename,".RData",sep="") 
print(filename)
if(FALSE) 
	{
	# Set the switch to TRUE for the first time. For experimentation with subsequent runs, read from saved files.		
	univ_results = UNIV_DECON(ws[,jj],mis,5,5,K.t,simsize_univ,burnin_univ,TRUE) 
	save(file=filename,univ_results)
	}
load(filename)
knots[,jj] = univ_results$knots			# The knot points that define the B-splines
xs[,jj] = univ_results$xs				# The initial estimates of X
thetas[,jj] = univ_results$thetas		# The log cofficients of the B-splines
es[,jj] = univ_results$es				# The initial scaled residuals
}
xs.current = xs.proposed = xs


Var.xs = function(xs,knots,thetas)
	{
	xs = as.matrix(xs)
	knots = as.matrix(knots)
	thetas = as.matrix(thetas)
	y = matrix(0,nrow=nrow(xs),ncol=ncol(xs))
	for(jj in 1:ncol(as.matrix(y)))
		y[,jj] = B.basis(xs[,jj],knots[,jj])%*%exp(thetas[,jj])
	return(y)
	}


# In the following, 
# K denotes the number of mixture components, 
# K.star denotes the number of latent factors,
# Pi denotes the mixture probabilities, C denotes the cluster labels, 
# Mu denotes the cluster mean, Lambda denotes the factor loading matrices, 
# Eta denotes the latent factors, Sigma denotes the common diagonal components,
# alpha denotes the hyper-parameter of the Dirichlet prior on the mixture probabilities,
# Mu0 and Sigma0 denote hyper-parameters of multivariate normal priors on Mu,
# a.sigma.sq and a.sigma.sq denote hyper-parameters of inverse gamma priors on the elements of Sigma,
# a.delta controls the shrinkage as the column index increases - a.delta=1 means no shrinkage and a.delta>1 with increaing values implying more shrinkage, 
# phi controls the local shrinkage of the factor loadings.


### For the density of X ###

d_clust = Mclust(as.matrix(xs), G=1:5)
m.best = length(unique(d_clust$class))
cat("Model-based optimal number of clusters:", m.best, "\n")
K.xs = m.best+3
Pi.xs = rep(0,K.xs)
Pi.xs[1:m.best] = c(d_clust$parameters$pro)
C.xs = d_clust$classification
Mu.xs = matrix(rep(colMeans(xs),K.xs),nrow=K.xs,ncol=d,byrow=T)
Mu.xs[1:m.best,] = t(d_clust$parameters$mean)

K.star.xs = (d+1)%/%2

alpha.xs = 1

a.sigma.sq.xs = 1
b.sigma.sq.xs = 1
tau.xs = matrix(1,nrow=K.xs,ncol=K.star.xs)
phi.xs = matrix(1,nrow=K.xs*d,ncol=K.star.xs)
nu.phi.xs = rep(1,K.xs)
delta.xs = matrix(1,nrow=K.xs,ncol=K.star.xs)
a.delta.xs = c(1,rep(1.4,K.star.xs-1))

sigma.sq.xs = rep(1,d)
Sigma.xs = diag(sigma.sq.xs)
Lambda.xs = matrix(0,nrow=K.xs*d,ncol=K.star.xs)
Eta.xs = matrix(0,nrow=n,ncol=K.star.xs) 

Mu0.xs = colMeans(xs)
Sigma0.xs = diag(apply(xs,2,"var"))


### For the density of the scaled errors ###

d_clust = Mclust(as.matrix(es), G=1:5)
m.best = length(unique(d_clust$class))
cat("Model-based optimal number of clusters:", m.best, "\n")
K.es = m.best+3
Pi.es = rep(0,K.es)
Pi.es[1:m.best] = c(d_clust$parameters$pro)
C.es = d_clust$classification
Mu.es = matrix(rep(colMeans(es),K.es),nrow=K.es,ncol=d,byrow=T)
Mu.es[1:m.best,] = t(d_clust$parameters$mean)

K.star.es = (d+1)%/%2

alpha.es = 1

a.sigma.sq.es = 1
b.sigma.sq.es = 1
tau.es = matrix(1,nrow=K.es,ncol=K.star.es)
phi.es = matrix(1,nrow=K.es*d,ncol=K.star.es)
nu.phi.es = rep(1,K.es)
delta.es = matrix(1,nrow=K.es,ncol=K.star.es)
a.delta.es = c(1,rep(1.4,K.star.es-1))

sigma.sq.es = rep(1,d)
Sigma.es = diag(sigma.sq.es)
Lambda.es = matrix(0,nrow=K.es*d,ncol=K.star.es)
Eta.es = matrix(0,nrow=n.es,ncol=K.star.es) 

Mu0.es = rep(0,d)
Sigma0.es = diag(rep(2,d))





###########################################
### MCMC Storage and Proposal Variables ###
###########################################

d.ordinates.xs = matrix(0,nrow=n.xs,ncol=K.xs)
d.ordinates.es = matrix(0,nrow=n.es,ncol=K.es)

min.xs = apply(apply(xs,2,"range"),2,"min")
max.xs = apply(apply(xs,2,"range"),2,"max")
Sigma.xs.proposal = diag(0.1*(max.xs-min.xs)/6)

current.likelihood.temp = proposed.likelihood.temp = numeric(N)
prior.current.xs = prior.proposed.xs = rep(1,n)
	
acc.thetas = numeric(d)

S.xs = S.xs.current = S.xs.proposed = matrix(1,nrow=n,ncol=d)
postmean.xs.mat = matrix(0,nrow=n,ncol=d)

CovMat.xs = array(0,dim=c(K.xs,d,d))
CovMat.es = array(0,dim=c(K.es,d,d))

xs.MCMC 		= array(0,dim=c(n,d,simsize))					

Pi.xs.MCMC		= array(0,dim=c(simsize-burnin,K.xs))					
Mu.xs.MCMC		= array(0,dim=c(K.xs,d,simsize-burnin))				
Lambda.xs.MCMC	= array(0,dim=c(K.xs*d,K.star.xs,simsize-burnin))				
sigma.sq.xs.MCMC= array(0,dim=c(simsize-burnin,d))			

Pi.es.MCMC	 	= array(0,dim=c(simsize-burnin,K.es))	
Mu.es.MCMC 		= array(0,dim=c(K.es,d,simsize-burnin))		
Lambda.es.MCMC 	= array(0,dim=c(K.es*d,K.star.es,simsize-burnin))		
sigma.sq.es.MCMC= array(0,dim=c(simsize-burnin,d))		

K.xs.MCMC 		= numeric(simsize)
K.es.MCMC 		= numeric(simsize)





	##################
	### Start MCMC ###
	##################
	

for(iii in 1:simsize)
	{
	if(iii%%10==0)
		print(iii)
			
			
	### Updating the parameters of the Distribution of X	
		
	### Updating the cluster indices
	for(kk in 1:K.xs)
		d.ordinates.xs[,kk] = dmvnorm(xs,Mu.xs[kk,],Lambda.xs[((kk-1)*d+1):(kk*d),]%*%t(Lambda.xs[((kk-1)*d+1):(kk*d),])+Sigma.xs)
	for(ii in 1:n.xs)
		C.xs[ii] = sample(K.xs,1,prob=Pi.xs*d.ordinates.xs[ii,])
	
	
	### Updating cluster probabilities
	n.kk.xs = tabulate(C.xs,nbins=K.xs)
	Pi.xs = rdirichlet(1,alpha.xs/K.xs+n.kk.xs)	

	
	### Updating Mu.xs
	for(kk in 1:K.xs)
		{
		temp = which(C.xs==kk)
		Sigma0.xs.temp = solve(solve(Sigma0.xs) + length(temp) * solve(Sigma.xs))
		xs.temp = xs[temp,]-Eta.xs[temp,]%*%t(Lambda.xs[((kk-1)*d+1):(kk*d),])		# Takes care of the case n.kk.xs[kk]==1
		Mu0.xs.temp = Sigma0.xs.temp %*% (solve(Sigma0.xs) %*% Mu0.xs + solve(Sigma.xs) %*% colSums(xs.temp))
		Mu.xs[kk,] = rmvnorm(1,Mu0.xs.temp,Sigma0.xs.temp)
		}
		
		
	# Updating Lambda.xs
	for(kk in which(n.kk.xs>0))
		{
		temp = which(C.xs==kk)
		for(jj in 1:d)
			{
			xs.temp = xs[temp,jj] - rep(Mu.xs[kk,jj],n.kk.xs[kk])
			D.inv.temp = diag( phi.xs[((kk-1)*d+jj),] * tau.xs[kk,],K.star.xs,K.star.xs ) 
			Eta.xs.temp = Eta.xs[temp,]
			if(n.kk.xs[kk]==1)
				Eta.xs.temp = t(as.matrix(Eta.xs[temp,]))
			Eta.mat = t(Eta.xs.temp) %*% Eta.xs.temp
			Sigma0.lambda.temp = solve(D.inv.temp + Eta.mat/sigma.sq.xs[jj])
			Sigma0.lambda.temp = (Sigma0.lambda.temp + t(Sigma0.lambda.temp))/2 	# for numerical symmetry
			Mu0.lambda.temp = Sigma0.lambda.temp %*% t(Eta.xs.temp) %*% xs.temp / sigma.sq.xs[jj]
			Lambda.xs[((kk-1)*d+jj),] = rmvnorm(1,Mu0.lambda.temp,Sigma0.lambda.temp)			
			}
		}	
	for(kk in which(n.kk.xs<1))
		{
		for(jj in 1:d)
			{
			D.inv.temp = diag( phi.xs[((kk-1)*d+jj),] * tau.xs[kk,],K.star.xs,K.star.xs )
			Sigma0.lambda.temp = solve(make.positive.definite(D.inv.temp))
			Sigma0.lambda.temp = (Sigma0.lambda.temp + t(Sigma0.lambda.temp))/2 	# for numerical symmetry
			Lambda.xs[((kk-1)*d+jj),] = rmvnorm(1,rep(0,K.star.xs),Sigma0.lambda.temp)			
			}
		}
	Lambda.xs.new = Lambda.xs
	temp = which(abs(Lambda.xs)>5, arr.ind=TRUE)	
	Lambda.sign = sign(Lambda.xs)
	Lambda.xs.new[temp] = abs(rnorm(nrow(temp)))*Lambda.sign[temp]			
	Lambda.xs = Lambda.xs.new


		
	
	# Updating Eta.xs
	for(kk in which(n.kk.xs>0))
		{
		temp = which(C.xs==kk)
		Lambda.xs.temp = Lambda.xs[((kk-1)*d+1):(kk*d),]
		Sigma0.eta.temp = solve(diag(rep(1,K.star.xs)) + t(Lambda.xs.temp)%*%solve(Sigma.xs)%*%Lambda.xs.temp)
		xs.temp = t(xs[temp,]-matrix(rep(Mu.xs[kk,],n.kk.xs[kk]),nrow=n.kk.xs[kk],byrow=T))		# Takes care of the case n.kk.xs[kk]==1.
		Mu0.eta.temp = t(Sigma0.eta.temp %*% t(Lambda.xs.temp) %*% solve(Sigma.xs) %*% xs.temp)
		Eta.xs[temp,] = Mu0.eta.temp + rmvnorm(n.kk.xs[kk],rep(0,K.star.xs),Sigma0.eta.temp)
		}

	
	# Updating sigma.xs
	for(jj in 1:d)
		sigma.sq.xs[jj] = 1/rgamma(1,shape=a.sigma.sq.xs+n/2,rate=b.sigma.sq.xs + 0.5*sum((xs[,jj]-Mu.xs[C.xs,jj]-rowSums(Lambda.xs[((C.xs-1)*d+jj),]*Eta.xs))^2))
	Sigma.xs = diag(sigma.sq.xs)
	
		
	# Updating phi.xs		
	if((iii>min(300,burnin/2))&&(T))
	{		
	phi0.mat = Lambda.xs^2*matrix(rep(tau.xs,d),nrow=K.xs*d,byrow=F)
	phi.temp = rgamma(K.xs*d*K.star.xs,shape=(matrix(nu.phi.xs,nrow=K.xs*d,ncol=K.star.xs)+1)/2,rate=(matrix(nu.phi.xs,nrow=K.xs*d,ncol=K.star.xs)+phi0.mat)/2) # Tested
	phi.xs = matrix(phi.temp,nrow=K.xs*d,ncol=K.star.xs)
	}
	
	
	# Updating delta.xs (and tau.xs)
	if((iii>min(300,burnin/2))&&(T))
	{		
	mat.temp = phi.xs*Lambda.xs^2
	for(kk in 1:K.xs)
		for(hh in 1:K.star.xs)
			{
			rate.temp = 1
			for(mm in hh:K.star.xs)
				rate.temp = rate.temp + (prod(delta.xs[kk,1:mm])/delta.xs[kk,hh])*sum(mat.temp[((kk-1)*d+1):(kk*d),mm])/2 	# Works since mm >= hh
			delta.xs[kk,hh] = rgamma(1,shape=a.delta.xs[hh]+d*(K.star.xs-hh+1)/2, rate=rate.temp)
			tau.xs[kk,hh] = prod(delta.xs[kk,1:hh])
			}
	}	
		
			

	# Dump redundant components on the fly

	if(iii==iteration_no_to_change_on_the_fly)
	{
	keep = which(Pi.xs>0.05)
	K.xs =  length(keep)
	discard = setdiff(C.xs,keep)

	for(jj in discard)
		{
		temp = Inf
		keep.index = Inf
		temp.C.xs = which(C.xs==jj)
		for(kk in keep)
			{
			temp2 = sum(Mu.xs[jj,] - Mu.xs[kk,])^2
			if(temp2<temp)
				{
				temp = temp2
				keep.index = kk
				}
			}
		C.xs[temp.C.xs] = keep.index
		Mu.xs[jj,] = Mu.xs[keep.index,]
		Lambda.xs[((jj-1)*d+1):(jj*d),] = Lambda.xs[((keep.index-1)*d+1):(keep.index*d),]
		phi.xs[((jj-1)*d+1):(jj*d),] = phi.xs[((keep.index-1)*d+1):(keep.index*d),]
		tau.xs[jj,] = tau.xs[keep.index,]
		delta.xs[jj,] = delta.xs[keep.index,]
		}
	C.xs.new = numeric(n)
	temp.keep = keep
	jj = 0
	for(kk in keep)
		{
		jj = jj+1
		temp = which(C.xs==kk)
		C.xs.new[temp] = C.xs[temp] - min(temp.keep)+jj
		temp.keep = setdiff(temp.keep,kk)
		}
	Mu.xs.new = Mu.xs[keep,]
	Lambda.xs.new = phi.xs.new = matrix(NA,nrow=K.xs*d,ncol=K.star.xs)
	jj = 0
	for(kk in keep)
		{
		jj = jj+1
		Lambda.xs.new[((jj-1)*d+1):(jj*d),] = Lambda.xs[((kk-1)*d+1):(kk*d),]
		phi.xs.new[((jj-1)*d+1):(jj*d),] = phi.xs[((kk-1)*d+1):(kk*d),]
		}
	tau.xs.new = tau.xs[keep,]
	delta.xs.new = delta.xs[keep,]
	d.ordinates.xs.new = d.ordinates.xs[,keep]

	C.xs = C.xs.new
	Mu.xs = Mu.xs.new
	Lambda.xs = Lambda.xs.new
	phi.xs = phi.xs.new
	tau.xs = tau.xs.new
	delta.xs = delta.xs.new
	Pi.xs = Pi.xs[keep]/sum(Pi.xs[keep])
	n.kk.xs = tabulate(C.xs,nbins=K.xs)
	d.ordinates.xs = d.ordinates.xs.new
	nu.phi.xs = rep(1,K.xs)
	if(K.xs==1)
		{
		Mu.xs = t(as.matrix(Mu.xs.new))
		tau.xs = t(as.matrix(tau.xs.new))
		delta.xs = t(as.matrix(delta.xs.new))
		d.ordinates.xs = as.matrix(d.ordinates.xs.new)
		}

	Pi.xs.MCMC			= array(0,dim=c(simsize-burnin,K.xs))					
	Mu.xs.MCMC			= array(0,dim=c(K.xs,d,simsize-burnin))				
	Lambda.xs.MCMC		= array(0,dim=c(K.xs*d,K.star.xs,simsize-burnin))				
	sigma.sq.xs.MCMC	= array(0,dim=c(simsize-burnin,d))			
	}
	
	
	
	
	### Updating X
	
	for(kk.xs in 1:K.xs)
		CovMat.xs[kk.xs,,] = Lambda.xs[((kk.xs-1)*d+1):(kk.xs*d),]%*%t(Lambda.xs[((kk.xs-1)*d+1):(kk.xs*d),]) + Sigma.xs
	for(kk.es in 1:K.es)
		CovMat.es[kk.es,,] = Lambda.es[((kk.es-1)*d+1):(kk.es*d),]%*%t(Lambda.es[((kk.es-1)*d+1):(kk.es*d),])+Sigma.es

	if(iii>min(30,burnin/2))
	{
	xs.proposed = xs.current + rmvnorm(n,rep(0,d),Sigma.xs.proposal)
	S.xs.current = sqrt(Var.xs(xs.current,knots,thetas))
	S.xs.proposed = sqrt(Var.xs(xs.proposed,knots,thetas))
	S.xs.proposed[S.xs.proposed<0.0000001] = 0.0001

	for(ii in 1:n)
		{
		kk.xs = C.xs[ii]

		prior.current.xs[ii] = dmvnorm(xs.current[ii,],Mu.xs[kk.xs,],CovMat.xs[kk.xs,,])
		prior.proposed.xs[ii] = dmvnorm(xs.proposed[ii,],Mu.xs[kk.xs,],CovMat.xs[kk.xs,,])
			
		for(jj in inds1[ii]:inds2[ii])
			{
			kk.es = C.es[jj]
			
			Covmat.temp.current = diag(S.xs.current[ii,]) %*% CovMat.es[kk.es,,] %*% diag(S.xs.current[ii,])
			current.likelihood.temp[jj] = dmvnorm(ws[jj,],xs.current[ii,]+S.xs.current[ii,]*Mu.es[kk.es,],Covmat.temp.current)
			
			Covmat.temp.proposed = diag(S.xs.proposed[ii,]) %*% CovMat.es[kk.es,,] %*% diag(S.xs.proposed[ii,])
			proposed.likelihood.temp[jj] = dmvnorm(ws[jj,],xs.proposed[ii,]+S.xs.proposed[ii,]*Mu.es[kk.es,],Covmat.temp.proposed)
			}
		}
		
	current.likelihood = tapply(current.likelihood.temp,inds,"prod")
	proposed.likelihood = tapply(proposed.likelihood.temp,inds,"prod")
	
	mh.ratio = (prior.proposed.xs*proposed.likelihood)/(prior.current.xs*current.likelihood)
	mh.ratio[which(is.nan(mh.ratio)==T)] = 0
	acc.prob = runif(n)
	inds.to.replace = (1:n)[acc.prob<mh.ratio]
	xs[inds.to.replace,] = xs.current[inds.to.replace,] = xs.proposed[inds.to.replace,]
	us = (ws-xs[inds,])
	if((iii>0.8*burnin)&&(T))
	{S.xs = sqrt(Var.xs(xs[inds,],knots,thetas))
	es = us/S.xs}
	}						


	
	### Updating the parameters of the Distribution of measurement errors		
		
	### Updating the cluster indices
	for(kk in 1:K.es)
		d.ordinates.es[,kk] = dmvnorm(es,Mu.es[kk,],Lambda.es[((kk-1)*d+1):(kk*d),]%*%t(Lambda.es[((kk-1)*d+1):(kk*d),])+Sigma.es)
	for(ii in 1:n.es)
		C.es[ii] = sample(K.es,1,prob=Pi.es*d.ordinates.es[ii,])
	
	
	### Updating cluster probabilities
	n.kk.es = tabulate(C.es,nbins=K.es)
	Pi.es = rdirichlet(1,alpha.es/K.es+n.kk.es)	

	
	### Updating Mu.es
	if(K.es==1)
		{Mu.es = matrix(0,nrow=1,ncol=d)}else
	{Big.Mu0.es = numeric((K.es+1)*d)
	Big.Sigma0.es = matrix(0,nrow=(K.es+1)*d,ncol=(K.es+1)*d)
	for(kk in 1:K.es)
		{
		temp = which(C.es==kk)
		Sigma0.es.temp = solve(solve(Sigma0.es) + length(temp) * solve(Sigma.es))
		es.temp = es[temp,]-Eta.es[temp,]%*%t(Lambda.es[((kk-1)*d+1):(kk*d),])		# Takes care of the case n.kk.es[kk]==1
		Mu0.es.temp = Sigma0.es.temp %*% (solve(Sigma0.es) %*% Mu0.es + solve(Sigma.es) %*% colSums(es.temp))
		Big.Mu0.es[((kk-1)*d+1):(kk*d)] = Mu0.es.temp
		Big.Sigma0.es[((kk-1)*d+1):(kk*d),((kk-1)*d+1):(kk*d)] = Sigma0.es.temp
	
		Big.Mu0.es[(K.es*d+1):((K.es+1)*d)] = Big.Mu0.es[(K.es*d+1):((K.es+1)*d)] + Pi.es[kk] * Mu0.es.temp
		Big.Sigma0.es[(K.es*d+1):((K.es+1)*d),(K.es*d+1):((K.es+1)*d)] = Big.Sigma0.es[(K.es*d+1):((K.es+1)*d),(K.es*d+1):((K.es+1)*d)] + Pi.es[kk]^2 * Sigma0.es.temp
		Big.Sigma0.es[((kk-1)*d+1):(kk*d),(K.es*d+1):((K.es+1)*d)] = Big.Sigma0.es[(K.es*d+1):((K.es+1)*d),((kk-1)*d+1):(kk*d)] = Pi.es[kk] * Sigma0.es.temp
		}
		
	Big.Mu0.es1 = Big.Mu0.es[1:(K.es*d)] - Big.Sigma0.es[1:(K.es*d),(K.es*d+1):((K.es+1)*d)] %*% solve(Big.Sigma0.es[(K.es*d+1):((K.es+1)*d),(K.es*d+1):((K.es+1)*d)]) %*% Big.Mu0.es[(K.es*d+1):((K.es+1)*d)] 
	Big.Sigma0.es11 = make.positive.definite(Big.Sigma0.es[1:(K.es*d),1:(K.es*d)] - Big.Sigma0.es[1:(K.es*d),(K.es*d+1):((K.es+1)*d)] %*% solve(Big.Sigma0.es[(K.es*d+1):((K.es+1)*d),(K.es*d+1):((K.es+1)*d)]) %*% Big.Sigma0.es[(K.es*d+1):((K.es+1)*d),1:(K.es*d)])

	Mu.es[1:(K.es-1),] = matrix(rmvnorm(1,Big.Mu0.es1[1:((K.es-1)*d)],Big.Sigma0.es11[1:((K.es-1)*d),1:((K.es-1)*d)]),nrow=(K.es-1),byrow=T)
	if(K.es>2)
		{Mu.es[K.es,] = - colSums(Pi.es[1:(K.es-1)]* Mu.es[1:(K.es-1),]) / Pi.es[K.es]}else
		{Mu.es[K.es,] = - Pi.es[1:(K.es-1)]* Mu.es[1:(K.es-1),] / Pi.es[K.es]}
	}
	
	
	# Updating Lambda.es
	for(kk in which(n.kk.es>0))
		{
		temp = which(C.es==kk)
		for(jj in 1:d)
			{
			es.temp = es[temp,jj] - rep(Mu.es[kk,jj],n.kk.es[kk])
			D.inv.temp = diag( phi.es[((kk-1)*d+jj),] * tau.es[kk,],K.star.es,K.star.es )
			Eta.es.temp = Eta.es[temp,]
			if(n.kk.es[kk]==1)
				Eta.es.temp = t(as.matrix(Eta.es[temp,]))
			Eta.mat = t(Eta.es.temp) %*% Eta.es.temp
			Sigma0.lambda.temp = solve(D.inv.temp + Eta.mat/sigma.sq.es[jj])
			Sigma0.lambda.temp = (Sigma0.lambda.temp + t(Sigma0.lambda.temp))/2 	# for numerical symmetry
			Mu0.lambda.temp = Sigma0.lambda.temp %*% t(Eta.es.temp) %*% es.temp / sigma.sq.es[jj]
			Lambda.es[((kk-1)*d+jj),] = rmvnorm(1,Mu0.lambda.temp,Sigma0.lambda.temp)			
			}
		}	
	for(kk in which(n.kk.es<1))
		{
		for(jj in 1:d)
			{
			D.inv.temp = diag( phi.es[((kk-1)*d+jj),] * tau.es[kk,],K.star.es,K.star.es )
			Sigma0.lambda.temp = solve(make.positive.definite(D.inv.temp))
			Sigma0.lambda.temp = (Sigma0.lambda.temp + t(Sigma0.lambda.temp))/2 	# for numerical symmetry
			Lambda.es[((kk-1)*d+jj),] = rmvnorm(1,rep(0,K.star.es),Sigma0.lambda.temp)			
			}
		}	
	Lambda.es.new = Lambda.es
	temp = which(abs(Lambda.es)>5, arr.ind=TRUE)	
	Lambda.sign = sign(Lambda.es)
	Lambda.es.new[temp] = abs(rnorm(nrow(temp)))*Lambda.sign[temp]			
	Lambda.es = Lambda.es.new


		
	# Updating Eta.es
	for(kk in which(n.kk.es>0))
		{
		temp = which(C.es==kk)
		Lambda.es.temp = Lambda.es[((kk-1)*d+1):(kk*d),]
		Sigma0.eta.temp = solve(diag(rep(1,K.star.es)) + t(Lambda.es.temp)%*%solve(Sigma.es)%*%Lambda.es.temp)
		es.temp = t(es[temp,]-matrix(rep(Mu.es[kk,],n.kk.es[kk]),nrow=n.kk.es[kk],byrow=T))	# Takes care of the case n.kk.es[kk]==1.
		Mu0.eta.temp = t(Sigma0.eta.temp %*% t(Lambda.es.temp) %*% solve(Sigma.es) %*% es.temp)
		Eta.es[temp,] = Mu0.eta.temp + rmvnorm(n.kk.es[kk],rep(0,K.star.es),Sigma0.eta.temp)
		}

	
	# Updating sigma.sq.es
	for(jj in 1:d)
		sigma.sq.es[jj] = 1/rgamma(1,shape=a.sigma.sq.es+N/2,rate=b.sigma.sq.es + 0.5*sum((es[,jj]-Mu.es[C.es,jj]-rowSums(Lambda.es[((C.es-1)*d+jj),]*Eta.es))^2))
	Sigma.es = diag(sigma.sq.es)	
	
	
	
		
	# Updating phi.es		
	if((iii>min(300,burnin/2))&&(T))
	{		
	phi0.mat = Lambda.es^2*matrix(rep(tau.es,d),nrow=K.es*d,byrow=F)
	phi.temp = rgamma(K.es*d*K.star.es,shape=(matrix(nu.phi.es,nrow=K.es*d,ncol=K.star.es)+1)/2,rate=(matrix(nu.phi.es,nrow=K.es*d,ncol=K.star.es)+phi0.mat)/2) # Tested
	phi.es = matrix(phi.temp,nrow=K.es*d,ncol=K.star.es)
	}
	
	
	# Updating delta.es (and tau.es)
	if((iii>min(300,burnin/2))&&(T))
	{		
	mat.temp = phi.es*Lambda.es^2
	for(kk in 1:K.es)
		for(hh in 1:K.star.es)
			{
			rate.temp = 1
			for(mm in hh:K.star.es)
				rate.temp = rate.temp + (prod(delta.es[kk,1:mm])/delta.es[kk,hh])*sum(mat.temp[((kk-1)*d+1):(kk*d),mm])/2 	# Works since mm >= hh
			delta.es[kk,hh] = rgamma(1,shape=a.delta.es[hh]+d*(K.star.es-hh+1)/2, rate=rate.temp)
			tau.es[kk,hh] = prod(delta.es[kk,1:hh])
			}
	}




	# Dump redundant components on the fly

	if(iii==iteration_no_to_change_on_the_fly)
	{
	keep = which(Pi.es>0.05)
	K.es =  length(keep)
	discard= setdiff(C.es,keep)

	for(jj in discard)
		{
		temp = Inf
		keep.index = Inf
		temp.C.es = which(C.es==jj)
		for(kk in keep)
			{
			temp2 = sum(Mu.es[jj,] - Mu.es[kk,])^2
			if(temp2<temp)
				{
				temp = temp2
				keep.index = kk
				}
			}
		C.es[temp.C.es] = keep.index
		Mu.es[jj,] = Mu.es[keep.index,]
		Lambda.es[((jj-1)*d+1):(jj*d),] = Lambda.es[((keep.index-1)*d+1):(keep.index*d),]
		phi.es[((jj-1)*d+1):(jj*d),] = phi.es[((keep.index-1)*d+1):(keep.index*d),]
		tau.es[jj,] = tau.es[keep.index,]
		delta.es[jj,] = delta.es[keep.index,]
		}
	C.es.new = numeric(N)
	temp.keep = keep
	jj = 0
	for(kk in keep)
		{
		jj = jj+1
		temp = which(C.es==kk)
		C.es.new[temp] = C.es[temp] - min(temp.keep)+jj
		temp.keep = setdiff(temp.keep,kk)
		}
	Mu.es.new = Mu.es[keep,]
	Lambda.es.new = phi.es.new = matrix(NA,nrow=K.es*d,ncol=K.star.es)
	jj = 0
	for(kk in keep)
		{
		jj = jj+1
		Lambda.es.new[((jj-1)*d+1):(jj*d),] = Lambda.es[((kk-1)*d+1):(kk*d),]
		phi.es.new[((jj-1)*d+1):(jj*d),] = phi.es[((kk-1)*d+1):(kk*d),]
		}
	tau.es.new = tau.es[keep,]
	delta.es.new = delta.es[keep,]
	d.ordinates.es.new = d.ordinates.es[,keep]

	C.es = C.es.new
	Mu.es = Mu.es.new
	Lambda.es = Lambda.es.new
	phi.es = phi.es.new
	tau.es = tau.es.new
	delta.es = delta.es.new
	Pi.es = Pi.es[keep]/sum(Pi.es[keep])
	n.kk.es = tabulate(C.es,nbins=K.es)
	d.ordinates.es = d.ordinates.es.new
	nu.phi.es = rep(1,K.es)
	if(K.es==1)
		{
		Mu.es = t(as.matrix(Mu.es.new))
		tau.es = t(as.matrix(tau.es.new))
		delta.es = t(as.matrix(delta.es.new))
		d.ordinates.es = as.matrix(d.ordinates.es.new)
		}

	Pi.es.MCMC			= array(0,dim=c(simsize-burnin,K.es))					
	Mu.es.MCMC			= array(0,dim=c(K.es,d,simsize-burnin))				
	Lambda.es.MCMC		= array(0,dim=c(K.es*d,K.star.es,simsize-burnin))				
	sigma.sq.es.MCMC	= array(0,dim=c(simsize-burnin,d))			
	}

	


	# Store the MCMC Output			
	xs.MCMC[,,iii] = xs					
	K.xs.MCMC[iii] = length(which(Pi.xs>0.05))
	K.es.MCMC[iii] = length(which(Pi.es>0.05))

	if(iii>burnin)
	{
	Pi.xs.MCMC[iii-burnin,] = Pi.xs					
	Mu.xs.MCMC[,,iii-burnin] = Mu.xs				
	Lambda.xs.MCMC[,,iii-burnin] = Lambda.xs				
	sigma.sq.xs.MCMC[iii-burnin,] = sigma.sq.xs				

	Pi.es.MCMC[iii-burnin,] = Pi.es	
	Mu.es.MCMC[,,iii-burnin] = Mu.es				
	Lambda.es.MCMC[,,iii-burnin] = Lambda.es				
	sigma.sq.es.MCMC[iii-burnin,] = sigma.sq.es			
	}
	
	}


xs.est = apply(xs.MCMC[,,(burnin+1):simsize],c(1,2),mean)
us.est = (ws-xs.est[inds,])
S.xs.est = sqrt(Var.xs(xs.est[inds,],knots,thetas))
es.est = us.est/S.xs.est




	######################
	### Create Figures ###
	######################

xs.grid.length = 100
xs.grid.lim.lwr = min(apply(Mu.xs,2,min))-3
xs.grid.lim.upr = max(apply(Mu.xs,2,max))+3
xs.grid = seq(xs.grid.lim.lwr,xs.grid.lim.upr,len=xs.grid.length)
xs.grid.mat = cbind(rep(xs.grid,times=rep(xs.grid.length,xs.grid.length)),rep(xs.grid,times=xs.grid.length))
density.xs.est.univ = matrix(0,nrow=xs.grid.length,ncol=d)
density.xs.est.array = array(0,dim=c(d,d,xs.grid.length,xs.grid.length))
sum_set = (simsize-burnin-min(25,simsize-burnin)+1):(simsize-burnin)
 

for(mm in 1:d)
	{		
	for(kk in 1:K.xs)
		for(iii in sum_set)
			{
			CovMat = Lambda.xs.MCMC[((kk-1)*d+1):(kk*d),,iii]%*%t(Lambda.xs.MCMC[((kk-1)*d+1):(kk*d),,iii]) + diag(sigma.sq.xs.MCMC[iii,])
			density.xs.est.univ[,mm] = density.xs.est.univ[,mm] + Pi.xs.MCMC[iii,kk] * dnorm(xs.grid,Mu.xs.MCMC[kk,mm,iii],sqrt(CovMat[mm,mm]))
			}
	density.xs.est.univ[,mm] = density.xs.est.univ[,mm]/length(sum_set)
	}
	
indices = combn(d,2)
for(ii in 1:ncol(indices))
	{
	dim.indices = indices[,ii]
	d1 = dim.indices[1]
	d2 = dim.indices[2]

	density.xs.est.vect = numeric(xs.grid.length^2)			
	for(kk in which(Pi.xs>0.05))
		for(iii in sum_set)
			{
			CovMat = Lambda.xs.MCMC[((kk-1)*d+1):(kk*d),,iii]%*%t(Lambda.xs.MCMC[((kk-1)*d+1):(kk*d),,iii]) + diag(sigma.sq.xs.MCMC[iii,])
			density.xs.est.vect = density.xs.est.vect + Pi.xs.MCMC[iii,kk] * dmvnorm(xs.grid.mat,Mu.xs.MCMC[kk,c(d1,d2),iii],CovMat[c(d1,d2),c(d1,d2)])
			}
	density.xs.est.vect = density.xs.est.vect/length(sum_set)
	density.xs.est.array[d1,d2,,] = matrix(density.xs.est.vect,nrow=xs.grid.length,byrow=T) 
	}





es.grid.length = 100
es.grid.lim.lwr = min(apply(Mu.es,2,min))-2
es.grid.lim.upr = max(apply(Mu.es,2,max))+2
es.grid.lim.lwr = -4
es.grid.lim.upr = 4
es.grid = seq(es.grid.lim.lwr,es.grid.lim.upr,len=es.grid.length)
es.grid.delta = (es.grid[2]-es.grid[1])
es.grid.mat = cbind(rep(es.grid,times=rep(es.grid.length,es.grid.length)),rep(es.grid,times=es.grid.length))
density.es.est.univ = matrix(0,nrow=es.grid.length,ncol=d)
density.es.est.array = array(0,dim=c(d,d,es.grid.length,es.grid.length))

	
for(mm in 1:d)
	{
	for(kk in 1:K.es)
		for(iii in sum_set)
			{
			CovMat = Lambda.es.MCMC[((kk-1)*d+1):(kk*d),,iii]%*%t(Lambda.es.MCMC[((kk-1)*d+1):(kk*d),,iii]) + diag(sigma.sq.es.MCMC[iii,])
			density.es.est.univ[,mm] = density.es.est.univ[,mm] + Pi.es.MCMC[iii,kk] * dnorm(es.grid,Mu.es.MCMC[kk,mm,iii],sqrt(CovMat[mm,mm]))
			}
	density.es.est.univ[,mm] = density.es.est.univ[,mm]/length(sum_set)
	}

indices = combn(d,2)
for(ii in 1:ncol(indices))
	{
	dim.indices = indices[,ii]
	d1 = dim.indices[1]
	d2 = dim.indices[2]
		
	density.es.est.vect = numeric(es.grid.length^2)
	for(kk in 1:K.es)
		for(iii in sum_set)
			{
			CovMat = Lambda.es.MCMC[((kk-1)*d+1):(kk*d),,iii]%*%t(Lambda.es.MCMC[((kk-1)*d+1):(kk*d),,iii]) + diag(sigma.sq.es.MCMC[iii,])
			density.es.est.vect = density.es.est.vect + Pi.es.MCMC[iii,kk] * dmvnorm(es.grid.mat,Mu.es.MCMC[kk,c(d1,d2),iii],CovMat[c(d1,d2),c(d1,d2)])
			}				
	density.es.est.vect = density.es.est.vect/length(sum_set)
	density.es.est.array[d1,d2,,] = matrix(density.es.est.vect,nrow=es.grid.length,byrow=T) 
	}







pdf("MVT_DECON_HTR_MLFA.pdf")

#######################
### 	XS	& ES	### 	Marginals of XS
#######################

laymat = diag(1:d)
laymat[lower.tri(laymat)] =  (d+1):(d*(d+1)/2)
laymat = t(laymat) 										# fixes the upper part
laymat_temp = diag(rep(0,d))
laymat_temp[upper.tri(laymat_temp)] = ((d*(d+1)/2)+1):d^2
laymat = laymat+t(laymat_temp)								# fixes the lower part
laymat

layout(laymat) 

par(mar=rep(1.85,4))
xs = data.frame(xs)	
my.cols = rev(brewer.pal(11, "RdYlBu"))
for(i in 1:d) 
	{
	plot(xs.grid,density.xs.est.univ[,i],col="blue",type="l",lwd=2)
	}
for(i in 1:(d-1))
	for(j in (i+1):d)
		{
		image(xs.grid,xs.grid,density.xs.est.array[i,j,,],col=terrain.colors(100))
		contour(xs.grid,xs.grid,density.xs.est.array[i,j,,],nlevels=45,drawlabels=FALSE,col=my.cols[6:11],add=T)
		legend("bottomright",legend=paste(i,j),cex=1,bty="n")
		} 
for(i in 2:d)
	for(j in 1:(i-1))
		{
		image(es.grid,es.grid,density.es.est.array[j,i,,],col=terrain.colors(100))
		contour(es.grid,es.grid,density.es.est.array[j,i,,],nlevels=45,drawlabels=FALSE,col=my.cols[6:11],add=T)
		legend("bottomright",legend=paste(j,i),cex=1,bty="n")
		} 


#######################
### 	XS	& ES	### 	Marginals of ES
#######################

laymat = diag(1:d)
laymat[lower.tri(laymat)] =  (d+1):(d*(d+1)/2)
laymat = t(laymat) 										# fixes the upper part
laymat_temp = diag(rep(0,d))
laymat_temp[upper.tri(laymat_temp)] = ((d*(d+1)/2)+1):d^2
laymat = laymat+t(laymat_temp)								# fixes the lower part
laymat

layout(laymat) 

par(mar=rep(1.85,4))
xs = data.frame(xs)	
my.cols = rev(brewer.pal(11, "RdYlBu"))
for(i in 1:d) 
	{
	plot(es.grid,density.es.est.univ[,i],col="blue",type="l",lwd=2)
	}
for(i in 1:(d-1))
	for(j in (i+1):d)
		{
		image(xs.grid,xs.grid,density.xs.est.array[i,j,,],col=terrain.colors(100))
		contour(xs.grid,xs.grid,density.xs.est.array[i,j,,],nlevels=45,drawlabels=FALSE,col=my.cols[6:11],add=T)
		legend("bottomright",legend=paste(i,j),cex=1,bty="n")
		} 
for(i in 2:d)
	for(j in 1:(i-1))
		{
		image(es.grid,es.grid,density.es.est.array[j,i,,],col=terrain.colors(100))
		contour(es.grid,es.grid,density.es.est.array[j,i,,],nlevels=45,drawlabels=FALSE,col=my.cols[6:11],add=T)
		legend("bottomright",legend=paste(j,i),cex=1,bty="n")
		} 


#################################
### 	VARIANCE FUNCTIONS    ###
#################################

par(mfrow=c(2,2))
par(mar=c(rep(2,4)))
for(jj in 1:d)
	{	
	xlimlwr = max(min(xs[,jj]),xs.grid[1])
	xlimupr = min(max(xs[,jj]),xs.grid[90])	
	s2is = as.vector(tapply(ws[,jj],inds,var))
	plot(xs[,jj],s2is,xlim=c(xlimlwr,xlimupr),ylim=c(0,30),xlab="X",ylab="Var Function",pch=16,col="gray60",cex=0.5)
	points(xs.grid,B.basis(xs.grid,knots[,jj])%*%exp(thetas[,jj]),type="l",lty=1,col="blue",lwd=2)
	}
par(mfrow=c(1,1))
dev.off()









###########################
### 	SAVE RESULTS    ###
###########################

save(ws, inds, mis, d, n, N, simsize_univ, burnin_univ, knots, thetas,
	xs.MCMC, K.xs.MCMC,	K.es.MCMC, 
	Pi.xs.MCMC, Mu.xs.MCMC,	Lambda.xs.MCMC,	sigma.sq.xs.MCMC,
	Pi.es.MCMC,	Mu.es.MCMC,	Lambda.es.MCMC,	sigma.sq.es.MCMC,
	sum_set,
	xs.grid, density.xs.est.univ, density.xs.est.array,
	es.grid, density.es.est.univ, density.es.est.array, file="MVT_DECON_HTR_MLFA.RData")
